//********************************************************************
UNIT dglRTTI;
//********************************************************************
INTERFACE

USES
   SysUtils, Classes, TypInfo, Contnrs;
//********************************************************************
TYPE
   ERTTIException = Class(Exception);

   TRTTI = Class;

   TRTTIClass = Class Of TRTTI;

   TRTTI_Register = Class(TObject)
      fClassName: String;
      fClassRef: TRTTIClass;
   End;

   TRTTIValue = Record
      TypeKind: TTypeKind;

      VInteger: LongInt;    // tkInteger
      VFloat: Extended;     // tkFloat
      VOrdinal: Word;       // tkEnumeration
      VString: String;      // tkString,tkLString
      VRTTIObject: TRTTI;   // tkClass
      VMethod: TMethod;     // tkMethod
   End;

   TDataExchange = (dxBinary, dxString);

//********************************************************************
// TRTTI provides information about its published properties
// (structure and content)

   TRTTI = Class(TObject)
   Private
      fPropCount   : Integer;
      fPropDataExchange : TDataExchange;
      fPropIndex   : Integer;
      fPropInfo    : pPropInfo;
      fPropList    : pPropList;
      fPropName    : String;
      fPropValue   : TRTTIValue;

      Procedure SetPropContent(ARTTIValue: TRTTIValue);
      Procedure SetPropIndex(AIndex: Integer);
       Function ValidPropIndex(AIndex: Integer): Boolean;
   Public
      Property CurPropName: String read fPropName;
      Property CurPropValue: TRTTIValue
               Read fPropValue Write SetPropContent;
      Property PropertyCount: Integer read fPropCount;
      Property PropDataOutput: TDataExchange
               Read fPropDataExchange Write fPropDataExchange;
      Property PropertyIndex: Integer
               Read fPropIndex Write SetPropIndex;

      Constructor Create; Reintroduce;
      Destructor Destroy; Override;
      Procedure OpenRTTI;
      Procedure CloseRTTI;
   End;

//********************************************************************
Procedure RegisterDGLClass(AClassName: String;
                            AClassRef: TRTTIClass);
Function FindDGLClass(AClassName: String): TRTTIClass;
Procedure ClearRTTIValue(Var ARTTIValue: TRTTIValue);
//********************************************************************
IMPLEMENTATION
//********************************************************************
VAR
   RTTI_ClassList: TObjectList;
//********************************************************************
//********************************************************************
//   T R T T I
//********************************************************************
Constructor TRTTI.Create;
Begin
   Inherited Create;
   fPropDataExchange:= dxBinary;
End;
//********************************************************************
Destructor TRTTI.Destroy;
Begin
   Inherited Destroy;
End;
//********************************************************************
Procedure TRTTI.OpenRTTI;
Begin
   If ClassInfo = Nil
      Then ERTTIException.Create
          ('OpenRTTI: No class information available');

   fPropCount:= GetTypeData(ClassInfo).PropCount;
   GetMem(fPropList, fPropCount * SizeOf(Pointer));
   GetPropList(ClassInfo,fPropList);
End;
//********************************************************************
Procedure TRTTI.CloseRTTI;
Begin
   FreeMem(fPropList);
End;
//********************************************************************
Procedure TRTTI.SetPropContent(ARTTIValue: TRTTIValue);
Begin
   FPropValue:= ARTTIValue;

   If fPropDataExchange = dxString Then
      With fPropValue Do Case TypeKind Of

         tkInteger     : VInteger:= StrToInt(VString);
         tkFloat       : VFloat:= StrToFloat(VString);
         tkEnumeration : VOrdinal:= StrToInt(VString);
         tkClass       : VString:= '';
         tkMethod      : VString:= '';

      End; // Of Case

   With fPropValue Do Case TypeKind Of

      tkInteger     : SetOrdProp(Self,fPropInfo,VInteger);
      tkFloat       : SetFloatProp(Self,fPropInfo,VFloat);
      tkEnumeration : SetOrdProp(Self,fPropInfo,VOrdinal);
{$IFDEF FPC}
      tkSString,tkLString,tkAString: SetStrProp(Self,fPropInfo,VString);
{$ELSE}
      tkString,tkLString: SetStrProp(Self,fPropInfo,VString);
{$ENDIF}
      tkClass       : SetObjectProp(Self,fPropInfo,VRTTIObject);
      tkMethod      : SetMethodProp(Self,fPropInfo,VMethod);

      Else
         ERTTIException.Create
            ('SetPropContent: TypeKind Error');

   End; // Of Case
End;
//********************************************************************
Procedure TRTTI.SetPropIndex(AIndex: Integer);
Begin
   fPropName:= 'ERROR';
   ClearRTTIValue(fPropValue);

   If ValidPropIndex(AIndex) Then
   Begin
      fPropIndex := AIndex;
      fPropInfo  := fPropList[AIndex];
      fPropName  := fPropInfo.Name;
      fPropValue.TypeKind:= fPropInfo.PropType^.Kind;

      With fPropValue Do Case TypeKind Of

         tkInteger     : VInteger:= GetOrdProp(Self,fPropInfo);
         tkFloat       : VFloat:=  GetFloatProp(Self,fPropInfo);
         tkEnumeration : VOrdinal:= GetOrdProp(Self,fPropInfo);
{$IFDEF FPC}
         tkSString,tkLString,tkAString: VString:= GetStrProp(Self,fPropInfo);
{$ELSE}
         tkString,tkLString: VString:= GetStrProp(Self,fPropInfo);
{$ENDIF}
         tkClass       : VRTTIObject:= TRTTI(GetObjectProp(Self,fpropInfo,TRTTI));
         tkMethod      : VMethod:= GetMethodProp(Self,fPropInfo);

         Else
            ERTTIException.Create('SetPropIndex: TypeKind Error');

      End; // Of Case

      If fPropDataExchange = dxString Then
         With fPropValue Do Case TypeKind Of

            tkInteger     : VString:= IntToStr(VInteger);
            tkFloat       : VString:= FloatToStr(VFloat);
            tkEnumeration : VString:= IntToStr(VOrdinal);
            tkClass       : VString:= '';
            tkMethod      : VString:= '';

         End; // Of Case
   End;
End;
//********************************************************************
Function TRTTI.ValidPropIndex(AIndex: Integer): Boolean;
Begin
   If (AIndex >= 0) and (AIndex < fPropCount)
      Then Result:= True
      Else Result:= False;
End;
//********************************************************************
//********************************************************************
// RTTI - Data Register
//********************************************************************
Function FindDGLClass(AClassName: String): TRTTIClass;
Var
   LocalRegister: TRTTI_Register;
   I: Integer;
   Found: Boolean;
Begin
   Result:= Nil;

   With RTTI_ClassList Do If Count > 0
      Then Begin
         I:= 0;
         Found:= False;
         While (I < Count) and (not Found) Do
            Begin
               LocalRegister:= List[I];
               With LocalRegister Do If fClassName = AClassName
                 Then Begin
                   Found:= True;
                   Result:= fClassRef;
                 End
                 Else Inc(I); //==============>
            End;
      End;
 End;
//********************************************************************
Procedure RegisterDGLClass(AClassName: String;
                           AClassRef: TRTTIClass);
Var
   LocalRegister: TRTTI_Register;
Begin
   LocalRegister:= TRTTI_Register.Create;
   With LocalRegister Do Begin
      fClassName:= AClassName;
      fClassRef:= AClassRef;
   End;
   RTTI_ClassList.Add(LocalRegister);
End;
//********************************************************************
Procedure ClearRTTIValue(Var ARTTIValue: TRTTIValue);
Begin
   With ARTTIValue Do Begin
      TypeKind:= tkUnknown;

      VInteger:= 0;
      VFloat  := 0;
      VOrdinal:= 0;
      VString := '';
      VRTTIObject := Nil;
      With VMethod Do Begin
         Code:= Nil;
         Data:= Nil;
      End;
   End;
End;
//********************************************************************
INITIALIZATION
//********************************************************************
   RTTI_ClassList:= TObjectList.Create;

//********************************************************************
FINALIZATION
//********************************************************************
   RTTI_ClassList.Free;

//********************************************************************
//********************************************************************
End.
